from django.urls import path

from arq_admin.views import (
    AllJobListView, DeferredJobListView, JobAbortView, JobDetailView,
    QueuedJobListView, QueueListView, RunningJobListView,
)

app_name = 'arq_admin'
urlpatterns = [
    path('', QueueListView.as_view(), name='home'),
    path('queue/<str:queue_name>/', AllJobListView.as_view(), name='all_jobs'),
    path('queue/<str:queue_name>/queued/', QueuedJobListView.as_view(), name='queued_jobs'),
    path('queue/<str:queue_name>/running/', RunningJobListView.as_view(), name='running_jobs'),
    path('queue/<str:queue_name>/deferred/', DeferredJobListView.as_view(), name='deferred_jobs'),
    path('queue/<str:queue_name>/<str:job_id>/', JobDetailView.as_view(), name='job_detail'),
    path('queue/<str:queue_name>/<str:job_id>/abort', JobAbortView.as_view(), name='job_abort'),
]



from dataclasses import dataclass
from datetime import datetime
from typing import Any, Optional, Union

from arq.jobs import JobDef, JobResult, JobStatus


@dataclass
class JobInfo(JobDef):
    job_id: str

    success: bool = False
    result: Optional[Any] = None
    start_time: Optional[datetime] = None
    finish_time: Optional[datetime] = None

    status: JobStatus = JobStatus.queued

    @classmethod
    def from_base(cls, base_info: Union[JobDef, JobResult], job_id: str) -> 'JobInfo':
        kwargs = {
            'function': base_info.function,
            'args': base_info.args,
            'kwargs': base_info.kwargs,
            'job_try': base_info.job_try,
            'enqueue_time': base_info.enqueue_time,
            'score': base_info.score,
            'job_id': job_id,
        }
        if isinstance(base_info, JobResult):
            kwargs.update({
                'success': base_info.success,
                'result': base_info.result,
                'start_time': base_info.start_time,
                'finish_time': base_info.finish_time,
            })

        return cls(**kwargs)  # type: ignore



import warnings
from collections import defaultdict
from typing import Any, Dict

from arq.connections import RedisSettings
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

if not hasattr(settings, 'ARQ_QUEUES'):
    raise ImproperlyConfigured('You have to define ARQ_QUEUES in settings.py')

ARQ_QUEUES: Dict[str, RedisSettings] = settings.ARQ_QUEUES

if not all(isinstance(redis_settings, RedisSettings) for redis_settings in ARQ_QUEUES.values()):
    raise ImproperlyConfigured('All values of "ARQ_QUEUES" must be RedisSettings')

ARQ_DESERIALIZER = getattr(settings, 'ARQ_DESERIALIZER', None)
ARQ_DESERIALIZER_BY_QUEUE = getattr(settings, 'ARQ_DESERIALIZER_BY_QUEUE', {})

if ARQ_DESERIALIZER:
    warnings.warn('ARQ_DESERIALIZER is deprecated, use ARQ_DESERIALIZER_BY_QUEUE', DeprecationWarning)

    if ARQ_DESERIALIZER_BY_QUEUE:
        raise ImproperlyConfigured('You cannot define both ARQ_DESERIALIZER and ARQ_DESERIALIZER_BY_QUEUE')

    ARQ_DESERIALIZER_BY_QUEUE = defaultdict(lambda: ARQ_DESERIALIZER)

ARQ_JOB_ABORT_TIMEOUT = getattr(settings, 'ARQ_JOB_ABORT_TIMEOUT', 5)



import asyncio
from operator import attrgetter
from typing import Any, Dict, List, Optional

from arq.jobs import JobStatus
from django.contrib import admin, messages
from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpRequest, HttpResponse
from django.shortcuts import redirect
from django.utils.decorators import method_decorator
from django.views.generic import DetailView, ListView

from arq_admin.job import JobInfo
from arq_admin.queue import Queue, QueueStats
from arq_admin.settings import ARQ_QUEUES


@method_decorator(staff_member_required, name='dispatch')
class QueueListView(ListView):
    template_name = 'arq_admin/queues.html'

    def get_queryset(self) -> List[QueueStats]:
        result = asyncio.run(self._gather_queues())
        return result

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context.update(admin.site.each_context(self.request))

        return context

    @staticmethod
    async def _get_queue_stats(queue_name: str) -> QueueStats:
        async with Queue.from_name(queue_name) as queue:
            return await queue.get_stats()

    @classmethod
    async def _gather_queues(cls) -> List[QueueStats]:
        result = []  # pragma: no cover
        for name in ARQ_QUEUES.keys():  # pragma: no cover
            result.append(await cls._get_queue_stats(name))
        return result


@method_decorator(staff_member_required, name='dispatch')
class BaseJobListView(ListView):
    paginate_by = 100
    template_name = 'arq_admin/jobs.html'

    status: Optional[JobStatus] = None
    job_status_label: Optional[str] = None

    @property
    def job_status(self) -> str:
        if self.job_status_label:
            return self.job_status_label

        return self.status.value.capitalize() if self.status else 'Unknown'

    def get_queryset(self) -> List[JobInfo]:
        queue_name = self.kwargs['queue_name']  # pragma: no cover
        jobs = asyncio.run(self._get_queue_jobs(queue_name))
        return sorted(jobs, key=attrgetter('enqueue_time'))

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context.update({
            **admin.site.each_context(self.request),
            'queue_name': self.kwargs['queue_name'],
            'job_status': self.job_status,
        })

        return context

    async def _get_queue_jobs(self, queue_name: str) -> List[JobInfo]:
        async with Queue.from_name(queue_name) as queue:
            return await queue.get_jobs(status=self.status)


class AllJobListView(BaseJobListView):
    job_status_label = 'All'


class QueuedJobListView(BaseJobListView):
    status = JobStatus.queued


class RunningJobListView(BaseJobListView):
    status = JobStatus.in_progress


class DeferredJobListView(BaseJobListView):
    status = JobStatus.deferred


class JobDetailView(DetailView):
    template_name = 'arq_admin/job_detail.html'

    def get_object(self, queryset: Optional[Any] = None) -> JobInfo:
        return asyncio.run(self._get_job_info())

    def get_context_data(self, **kwargs: Any) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context['queue_name'] = self.kwargs['queue_name']

        return context

    async def _get_job_info(self) -> JobInfo:
        async with Queue.from_name(self.kwargs['queue_name']) as queue:
            return await queue.get_job_by_id(self.kwargs['job_id'])


class JobAbortView(JobDetailView):
    template_name = 'arq_admin/job_abort.html'

    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:  # pragma: nocover
        aborted = asyncio.run(self._abort_job())
        if aborted:
            messages.success(request, 'Job aborted successfully')
        elif aborted is None:
            messages.warning(request, 'An issue happened while aborting the job, but it may be aborted anyway')
        else:
            messages.error(request, 'Job was not aborted')

        return redirect('arq_admin:job_detail', queue_name=self.kwargs['queue_name'], job_id=self.kwargs['job_id'])

    async def _abort_job(self) -> Optional[bool]:
        async with Queue.from_name(self.kwargs['queue_name']) as queue:
            return await queue.abort_job(self.kwargs['job_id'])



from io import open
from typing import List

from setuptools import setup

import arq_admin


def get_readme() -> str:
    """read the contents of the README file"""
    with open('README.md', encoding='utf-8') as f:
        return f.read()


def get_requirements() -> List[str]:
    """read the contents of the requirements file"""
    with open('requirements.txt', encoding='utf-8') as f:
        return [line.replace('==', '>=') for line in f.readlines()]


setup(
    name='arq-django-admin',
    version=arq_admin.__version__,
    setup_requires=['better-setuptools-git-version'],
    install_requires=get_requirements(),
    tests_require=[],
    packages=['arq_admin'],
    package_data={'arq_admin': ['py.typed']},
    include_package_data=True,
    author='Slava Skvortsov',
    description='An app that provides Django admin for ARQ',
    long_description=get_readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/SlavaSkvortsov/arq-django-admin',
    zip_safe=True,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Monitoring',
    ],
)



import asyncio
from contextlib import suppress
from dataclasses import dataclass, field
from typing import Any, List, Optional, Set

from arq import ArqRedis
from arq.connections import RedisSettings, create_pool
from arq.constants import result_key_prefix
from arq.jobs import DeserializationError, Job as ArqJob, JobDef, JobStatus
from django.utils import timezone

from arq_admin import settings
from arq_admin.job import JobInfo


@dataclass
class QueueStats:
    name: str
    host: str
    port: int
    database: int

    queued_jobs: Optional[int] = None
    running_jobs: Optional[int] = None
    deferred_jobs: Optional[int] = None

    error: Optional[str] = None


@dataclass
class Queue:
    redis_settings: RedisSettings
    name: str
    _redis: ArqRedis = field(init=False, default=None)  # type: ignore

    async def __aenter__(self) -> 'Queue':
        self._redis = await create_pool(self.redis_settings)
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self._redis.close()

    @classmethod
    def from_name(cls, name: str) -> 'Queue':
        return cls(
            name=name,
            redis_settings=settings.ARQ_QUEUES[name],
        )

    async def get_jobs(self, status: Optional[JobStatus] = None) -> List[JobInfo]:
        job_ids = await self._get_job_ids()

        if status:
            job_ids_tuple = tuple(job_ids)
            statuses = await asyncio.gather(*[self._get_job_status(job_id) for job_id in job_ids_tuple])
            job_ids = {job_id for (job_id, job_status) in zip(job_ids_tuple, statuses) if job_status == status}

        jobs: List[JobInfo] = await asyncio.gather(*[self.get_job_by_id(job_id) for job_id in job_ids])

        return jobs

    async def get_stats(self) -> QueueStats:
        result = QueueStats(
            name=self.name,
            host=str(self.redis_settings.host),
            port=self.redis_settings.port,
            database=self.redis_settings.database,
        )
        try:
            job_ids = await self._get_job_ids()
            statuses = await asyncio.gather(*[self._get_job_status(job_id) for job_id in job_ids])
        except Exception as ex:  # noqa: B902
            result.error = str(ex)
        else:
            result.queued_jobs = len([status for status in statuses if status == JobStatus.queued])
            result.running_jobs = len([status for status in statuses if status == JobStatus.in_progress])
            result.deferred_jobs = len([status for status in statuses if status == JobStatus.deferred])

        return result

    async def get_job_by_id(self, job_id: str) -> JobInfo:
        arq_job = ArqJob(
            job_id=job_id,
            redis=self._redis,
            _queue_name=self.name,
            _deserializer=settings.ARQ_DESERIALIZER_BY_QUEUE.get(self.name),
        )

        unknown_function_msg = "Can't find job"
        base_info = None
        try:
            base_info = await arq_job.info()
        except DeserializationError:
            unknown_function_msg = "Unknown, can't deserialize"

        if not base_info:
            base_info = JobDef(
                function=unknown_function_msg,
                args=(),
                kwargs={},
                job_try=-1,
                enqueue_time=timezone.now().replace(year=2077),
                score=420,
            )

        job_info = JobInfo.from_base(base_info, job_id)
        job_info.status = await arq_job.status()

        return job_info

    async def abort_job(self, job_id: str) -> Optional[bool]:
        # None here means we are not sure if the job was aborted or not
        arq_job = ArqJob(
            job_id=job_id,
            redis=self._redis,
            _queue_name=self.name,
            _deserializer=settings.ARQ_DESERIALIZER_BY_QUEUE.get(self.name),
        )
        with suppress(asyncio.TimeoutError):
            return await arq_job.abort(timeout=settings.ARQ_JOB_ABORT_TIMEOUT)

        return None

    async def _get_job_status(self, job_id: str) -> JobStatus:
        arq_job = ArqJob(
            job_id=job_id,
            redis=self._redis,
            _queue_name=self.name,
            _deserializer=settings.ARQ_DESERIALIZER_BY_QUEUE.get(self.name),
        )
        return await arq_job.status()

    async def _get_job_ids(self) -> Set[str]:
        raw_job_ids = set(await self._redis.zrangebyscore(self.name, '-inf', 'inf'))
        result_keys = await self._redis.keys(f'{result_key_prefix}*')
        raw_job_ids |= {key[len(result_key_prefix):] for key in result_keys}

        return {job_id.decode('utf-8') if isinstance(job_id, bytes) else job_id for job_id in raw_job_ids}
